local L = LibStub("AceLocale-3.0"):GetLocale("FlightHUD", false)

-- Config options
FlightHUDConfig = { 
  type='group',
  name = L["FlightHUD"],
  desc = L["A Heads Up Display (HUD) that shows information while flying or riding"],
  icon = [[Interface\Addons\FlightHUD\icon]],
  args = {
    General = {
      order = 1,
      type = "group",
      name = L["General Options"],
      desc = L["General Options"],
      args = {
        RefreshRate = {
          order = 1,
          type = 'range',
          name = L["Refresh Interval"],
          desc = L["Refresh the HUD every X seconds. (Lower values might be CPU intensive)"],
          get = function ()
            return FlightHUD.db.profile.RefreshRate
          end,
          set = function (i, arg)
            FlightHUD.db.profile.RefreshRate = arg
          end,
          min = 0,
          max = 1,
          step = 0.01,
        },
        HUDAlpha = {
          order = 2,
          type = 'range',
          name = L["HUD Alpha"],
          desc = L["The transparency value of the HUD"],
          get = function ()
            return FlightHUD.db.profile.HUDAlpha
          end,
          set = function (i, arg)
            FlightHUD.db.profile.HUDAlpha = arg
            FlightHUD.CompassHeadingImage1:SetAlpha(arg)
            FlightHUD.CompassHeadingImage2:SetAlpha(arg)
            FlightHUD.PitchImage:SetAlpha(arg)
            FlightHUD.ArtificialHorizonImage:SetAlpha(arg)
            FlightHUD.HeadingCenterImage:SetAlpha(arg)
            FlightHUD.HeadingText:SetAlpha(arg)
            FlightHUD.PitchText:SetAlpha(arg)
            FlightHUD.CoordinatesText:SetAlpha(arg)
            FlightHUD.SpeedText:SetAlpha(arg)
          end,
          min = 0,
          max = 1,
          step = 0.05,
        },
        HUDBGAlpha = {
          order = 3,
          type = 'range',
          name = L["HUD BG Alpha"],
          desc = L["The transparency value of the HUD background"],
          get = function ()
            return FlightHUD.db.profile.HUDBGAlpha
          end,
          set = function (i, arg)
            FlightHUD.db.profile.HUDBGAlpha = arg
            FlightHUD.timage:SetAlpha(arg)
          end,
          min = 0,
          max = 1,
          step = 0.05,
        },
        spacer11 = {
          order = 4,
          type = "description",
          name = "",
        },
        LockPosition = {
          order = 5,
          type = 'toggle',
          name = L["Lock"],
          desc = L["Lock the position of the HUD"],
          get = function ()
            return FlightHUD.db.profile.LockPosition
          end,
          set = function (i, arg)
            FlightHUD.db.profile.LockPosition = arg
          end
        },
      },
    },
    ShowEvents = {
      order = 2,
      type = "group",
      name = L["Show Events"],
      desc = L["Events which dictate when the HUD and its elements are shown."],
      args = {
        spacer21 = {
          order = 1,
          type = "description",
          name = L["When to show the HUD"],
        },
        ShowWhenFlying = {
          order = 2,
          type = 'toggle',
          name = L["Flying"],
          desc = L["Show the HUD when mounted on a flying mount"],
          disabled = function ()
            return FlightHUD.db.profile.ShowAlways
          end,
          get = function ()
            return FlightHUD.db.profile.ShowWhenFlying
          end,
          set = function (i, arg)
            FlightHUD.db.profile.ShowWhenFlying = arg
            FlightHUD:UNIT_AURA ()
          end
        },
        ShowWhenMounted = {
          order = 3,
          type = 'toggle',
          name = L["Mounted"],
          desc = L["Show the HUD when mounted on a ground mount"],
          disabled = function ()
            return FlightHUD.db.profile.ShowAlways
          end,
          get = function ()
            return FlightHUD.db.profile.ShowWhenMounted
          end,
          set = function (i, arg)
            FlightHUD.db.profile.ShowWhenMounted = arg
            FlightHUD:UNIT_AURA ()
          end
        },
        ShowWhenSwimming = {
          order = 4,
          type = 'toggle',
          name = L["Swimming"],
          desc = L["Show the HUD when swimming"],
          disabled = function ()
            return FlightHUD.db.profile.ShowAlways
          end,
          get = function ()
            return FlightHUD.db.profile.ShowWhenSwimming
          end,
          set = function (i, arg)
            FlightHUD.db.profile.ShowWhenSwimming = arg
            FlightHUD:UNIT_AURA ()
          end
        },
        ShowAlways = {
          order = 5,
          type = 'toggle',
          name = L["Always"],
          desc = L["Always show the HUD"],
          get = function ()
            return FlightHUD.db.profile.ShowAlways
          end,
          set = function (i, arg)
            FlightHUD.db.profile.ShowAlways = arg
            FlightHUD:UNIT_AURA ()
          end
        },
        spacer22 = {
          order = 6,
          type = "description",
          name = L["When to show pitch"],
        },
        ShowPitchWhenFlying = {
          order = 7,
          type = 'toggle',
          name = L["Flying"],
          desc = L["Show pitch when mounted on a flying mount"],
          disabled = function ()
            return FlightHUD.db.profile.ShowPitchAlways
          end,
          get = function ()
            return FlightHUD.db.profile.ShowPitchWhenFlying
          end,
          set = function (i, arg)
            FlightHUD.db.profile.ShowPitchWhenFlying = arg
            FlightHUD:UNIT_AURA ()
          end
        },
        ShowPitchWhenMounted = {
          order = 8,
          type = 'toggle',
          name = L["Mounted"],
          desc = L["Show pitch when mounted on a ground mount"],
          disabled = function ()
            return FlightHUD.db.profile.ShowPitchAlways
          end,
          get = function ()
            return FlightHUD.db.profile.ShowPitchWhenMounted
          end,
          set = function (i, arg)
            FlightHUD.db.profile.ShowPitchWhenMounted = arg
            FlightHUD:UNIT_AURA ()
          end
        },
        ShowPitchWhenSwimming = {
          order = 9,
          type = 'toggle',
          name = L["Swimming"],
          desc = L["Show pitch when swimming"],
          disabled = function ()
            return FlightHUD.db.profile.ShowPitchAlways
          end,
          get = function ()
            return FlightHUD.db.profile.ShowPitchWhenSwimming
          end,
          set = function (i, arg)
            FlightHUD.db.profile.ShowPitchWhenSwimming = arg
            FlightHUD:UNIT_AURA ()
          end
        },
        ShowPitchAlways = {
          order = 10,
          type = 'toggle',
          name = L["Always"],
          desc = L["Always show pitch"],
          get = function ()
            return FlightHUD.db.profile.ShowPitchAlways
          end,
          set = function (i, arg)
            FlightHUD.db.profile.ShowPitchAlways = arg
            FlightHUD:UNIT_AURA ()
          end
        },
        spacer23 = {
          order = 11,
          type = "description",
          name = L["When to show compass"],
        },
        ShowCompassWhenFlying = {
          order = 12,
          type = 'toggle',
          name = L["Flying"],
          desc = L["Show compass when mounted on a flying mount"],
          disabled = function ()
            return FlightHUD.db.profile.ShowCompassAlways
          end,
          get = function ()
            return FlightHUD.db.profile.ShowCompassWhenFlying
          end,
          set = function (i, arg)
            FlightHUD.db.profile.ShowCompassWhenFlying = arg
            FlightHUD:UNIT_AURA ()
          end
        },
        ShowCompassWhenMounted = {
          order = 13,
          type = 'toggle',
          name = L["Mounted"],
          desc = L["Show compass when mounted on a ground mount"],
          disabled = function ()
            return FlightHUD.db.profile.ShowCompassAlways
          end,
          get = function ()
            return FlightHUD.db.profile.ShowCompassWhenMounted
          end,
          set = function (i, arg)
            FlightHUD.db.profile.ShowCompassWhenMounted = arg
            FlightHUD:UNIT_AURA ()
          end
        },
        ShowCompassWhenSwimming = {
          order = 14,
          type = 'toggle',
          name = L["Swimming"],
          desc = L["Always show compass"],
          disabled = function ()
            return FlightHUD.db.profile.ShowCompassAlways
          end,
          get = function ()
            return FlightHUD.db.profile.ShowCompassWhenSwimming
          end,
          set = function (i, arg)
            FlightHUD.db.profile.ShowCompassWhenSwimming = arg
            FlightHUD:UNIT_AURA ()
          end
        },
        ShowCompassAlways = {
          order = 15,
          type = 'toggle',
          name = L["Always"],
          desc = L["Always show compass"],
          get = function ()
            return FlightHUD.db.profile.ShowCompassAlways
          end,
          set = function (i, arg)
            FlightHUD.db.profile.ShowCompassAlways = arg
            FlightHUD:UNIT_AURA ()
          end
        },
      },
    },
    SpeedVars = {
      order = 3,
      type = "group",
      name = L["Speed Controls"],
      desc = L["Control how speed is displayed."],
      args = {
        SpeedType = {
          order = 1,
          type = 'select',
          name = L["Speed display"],
          desc = L["Show speed as yards per second or percentage."],
          get = function ()
            return FlightHUD.db.profile.SpeedType
          end,
          set = function (i, arg)
            FlightHUD.db.profile.SpeedType = arg
          end,
          disabled = function ()
            return not FlightHUD.db.profile.ShowSpeedText
          end,
          values = {
            BOTTOM = L["Percentage"],
            LEFT = L["Yards per second"],
            RIGHT = L["Both"],
          },
        },
        WhichSpeedType = {
          order = 2,
          type = 'select',
          name = L["Which speed"],
          desc = L["Show speed as true speed or horizontal speed."],
          get = function ()
            return FlightHUD.db.profile.WhichSpeedType
          end,
          set = function (i, arg)
            FlightHUD.db.profile.WhichSpeedType = arg
          end,
          disabled = function ()
            return not FlightHUD.db.profile.ShowSpeedText
          end,
          values = {
            BOTTOM = L["True speed"],
            LEFT = L["Horizontal speed"],
            RIGHT = L["Both"],
          },
        },
        SpeedPrecision = {
          order = 3,
          type = 'range',
          name = L["Speed Precision"],
          desc = L["The number of decimal points to display for speed."],
          get = function ()
            return FlightHUD.db.profile.SpeedPrecision
          end,
          set = function (i, arg)
            FlightHUD.db.profile.SpeedPrecision = arg
          end,
          min = 0,
          max = 2,
          step = 1,
        },
        spacer31 = {
          order = 4,
          type = "description",
          name = "",
        },
        ShowSpeedPrefix = {
          order = 5,
          type = 'toggle',
          name = L["Speed Prefix Text"],
          desc = L["Prepend a prefix to denote the speed type ([H]orizontal or [T]rue)"],
          get = function ()
            return FlightHUD.db.profile.ShowSpeedPrefix
          end,
          set = function (i, arg)
            FlightHUD.db.profile.ShowSpeedPrefix = arg
          end
        },
      },
    },
    Scaling = {
      order = 4,
      type = "group",
      name = L["Scaling"],
      desc = L["Options for scaling the HUD and its elements."],
      args = {
        FrameScale = {
          order = 1,
          type = 'range',
          name = L["Frame Scale"],
          desc = L["Scale the HUD frame"],
          get = function ()
            return FlightHUD.db.profile.FrameScale
          end,
          set = function (i, arg)
            FlightHUD.db.profile.FrameScale = arg
            FlightHUD.FlightHUDFrame:SetScale(arg)
          end,
          min = 0.25,
          max = 2,
          step = 0.25,
        },
        PitchGraphicsScale = {
          order = 2,
          type = 'range',
          name = L["Pitch Graphics Scale"],
          desc = L["Scale the HUD pitch graphic"],
          get = function ()
            return FlightHUD.db.profile.PitchGraphicsScale
          end,
          set = function (i, arg)
            FlightHUD.db.profile.PitchGraphicsScale = arg
          end,
          min = 0,
          max = 20,
          step = 0.5,
        },
        FontScale = {
          order = 3,
          type = 'range',
          name = L["Font Scale"],
          desc = L["Scale the HUD text fonts"],
          get = function ()
            return FlightHUD.db.profile.FontScale
          end,
          set = function (i, arg)
            FlightHUD.db.profile.FontScale = arg
            local font = FlightHUD.PitchText:GetFont()
            FlightHUD.PitchText:SetFont(font, arg)
            FlightHUD.HeadingText:SetFont(font, arg)
            FlightHUD.CoordinatesText:SetFont(font, arg)
            FlightHUD.SpeedText:SetFont(font, arg)
          end,
          min = 6,
          max = 30,
          step = 0.5,
        },
      },
    },
    TextOptions = {
      order = 5,
      type = "group",
      name = L["Text Options"],
      desc = L["Toggle the display of certain texts on and off."],
      args = {
        ShowHeadingText = {
          order = 1,
          type = 'toggle',
          name = L["Heading text"],
          desc = L["Show the exact heading text in the HUD"],
          get = function ()
            return FlightHUD.db.profile.ShowHeadingText
          end,
          set = function (i, arg)
            FlightHUD.db.profile.ShowHeadingText = arg
            if (arg) then
              FlightHUD.HeadingText:Show()
            else
              FlightHUD.HeadingText:Hide()
            end
          end
        },
        ShowPitchText = {
          order = 2,
          type = 'toggle',
          name = L["Pitch text"],
          desc = L["Show the exact pitch text in the HUD"],
          get = function ()
            return FlightHUD.db.profile.ShowPitchText
          end,
          set = function (i, arg)
            FlightHUD.db.profile.ShowPitchText = arg
            if (arg) then
              FlightHUD.PitchText:Show()
            else
              FlightHUD.PitchText:Hide()
            end
          end
        },
        ShowCoordinatesText = {
          order = 3,
          type = 'toggle',
          name = L["Map coordinates text"],
          desc = L["Show the map coordinates text in the HUD"],
          get = function ()
            return FlightHUD.db.profile.ShowCoordinatesText
          end,
          set = function (i, arg)
            FlightHUD.db.profile.ShowCoordinatesText = arg
            if (arg) then
              FlightHUD.CoordinatesText:Show()
            else
              FlightHUD.CoordinatesText:Hide()
            end
          end
        },
        ShowSpeedText = {
          order = 4,
          type = 'toggle',
          name = L["Player speed text"],
          desc = L["Show the player's speed text in the HUD"],
          get = function ()
            return FlightHUD.db.profile.ShowSpeedText
          end,
          set = function (i, arg)
            FlightHUD.db.profile.ShowSpeedText = arg
            if (arg) then
              FlightHUD.SpeedText:Show()
            else
              FlightHUD.SpeedText:Hide()
            end
          end
        },
      },
    },
    WaypointOptions = {
      order = 6,
      type = "group",
      name = L["Waypoints"],
      desc = L["Waypoint Options"],
      args = {
        WaypointDisplay = {
          order = 1,
          type = 'select',
          name = L["Waypoint Display"],
          desc = L["Choose to display waypoints in an ellipse or a rectangle."],
          get = function ()
            return FlightHUD.db.profile.WaypointDisplay
          end,
          set = function (i, arg)
            FlightHUD.db.profile.WaypointDisplay = "BOTTOM"
            FlightHUD:UNIT_AURA ()
          end,
          values = {
            BOTTOM = L["Ellipse"],
            LEFT = L["Rectangle"],
          },
        },
        WorldmapModifier = {
          order = 2,
          type = "select",
          name = L["Worldmap modifier"],
          desc = L["Set the modifier used when right-clicking on the world map to create a waypoint"],
          get = function ()
            return FlightHUD.db.profile.WorldmapModifier
          end,
          set = function (i, arg)
            FlightHUD.db.profile.WorldmapModifier = arg
          end,
          values = {
            ["A"] = L["Alt"],
            ["C"] = L["Ctrl"],
            ["S"] = L["Shift"],
            ["AC"] = L["Alt-Ctrl"],
            ["AS"] = L["Alt-Shift"],
            ["CS"] = L["Ctrl-Shift"],
            ["ACS"] = L["Alt-Ctrl-Shift"],
          },
        },
        RemoveWaypointDistance = {
          order = 3,
          type = 'range',
          name = L["Waypoint distance"],
          desc = L["The distance from the waypoint before it is removed if it's not sticky"],
          get = function ()
            return FlightHUD.db.profile.RemoveWaypointDistance
          end,
          set = function (i, arg)
            FlightHUD.db.profile.RemoveWaypointDistance = arg
          end,
          min = 1,
          max = 100,
          step = 1,
        },
        spacer61 = {
          order = 4,
          type = "description",
          name = "",
        },
        QuestHelper = {
          order = 5,
          type = 'toggle',
          name = L["QuestHelper Integration"],
          desc = L["Make the QuestHelper default quest always appear as a waypoint in the HUD"],
          get = function ()
            return FlightHUD.db.profile.QuestHelper
          end,
          set = function (i, arg)
            FlightHUD.db.profile.QuestHelper = arg
            if arg == true then
              FlightHUD.SetupQuestHelperCallback ()
            else
              FlightHUD.RemoveQuestHelperCallback ()
            end
          end
        },
      },
    },
    ResetOptions = {
      order = 7,
      type = "group",
      name = L["Reset"],
      desc = L["Options for resetting all other options."],
      args = {
        ResetPosition = {
          order = -11,
          type = "execute",
          confirm = true,
          confirmText = L["Really reset the position of the HUD?"],
          name = L["Reset Position"],
          desc = L["Reset the position of the HUD back to the center of the screen"],
          func = function ()
            FlightHUD.ResetPosition()
          end
        },
        ResetAll = {
          order = -5,
          type = "execute",
          confirm = true,
          confirmText = L["Really reset all FlightHUD settings?"],
          name = L["Reset all settings"],
          desc = L["Resets all FlightHUD settings back to the default"],
          func = function ()
            FlightHUD.ResetAll()
          end
        },
      },
    },
    --[===[@debug@
    DebugOptions = {
      order = 8,
      type = "group",
      name = L["Debug"],
      desc = L["Options for debugging various aspects of the addon."],
      args = {
        EnableDebugWaypoints = {
          order = 1,
          type = 'toggle',
          name = L["Debug waypoints"],
          desc = L["Set the debug flag for waypoint code"],
          get = function ()
            return FlightHUD.db.profile.EnableDebugWaypoints
          end,
          set = function (i, arg)
            FlightHUD.db.profile.EnableDebugWaypoints = arg
          end
        },
        spacer81 = {
          order = 2,
          type = "description",
          name = "",
        },
        EnableDebugQuestHelper = {
          order = 3,
          type = 'toggle',
          name = L["Debug QuestHelper"],
          desc = L["Set the debug flag for QuestHelper code"],
          get = function ()
            return FlightHUD.db.profile.EnableDebugQuestHelper
          end,
          set = function (i, arg)
            FlightHUD.db.profile.EnableDebugQuestHelper = arg
          end
        },
        spacer82 = {
          order = 4,
          type = "description",
          name = "",
        },
      },
    },
    --@end-debug@]===]
  },
}

-- Config defaults
FlightHUDDefaultConfigDB = {
  global = {
    ["*"] = {},
  },
  profile = {
    ShowWhenFlying = true,
    ShowWhenMounted = false,
    ShowWhenSwimming = false,
    ShowAlways = false,
    ShowPitchWhenFlying = true,
    ShowPitchWhenMounted = true,
    ShowPitchWhenSwimming = true,
    ShowPitchAlways = false,
    ShowCompassWhenFlying = true,
    ShowCompassWhenMounted = true,
    ShowCompassWhenSwimming = true,
    ShowCompassAlways = false,
    SpeedType = "RIGHT",
    WhichSpeedType = "RIGHT",
    WaypointDisplay = "BOTTOM",
    RefreshRate = 0.1,
    HUDAlpha = 1,
    HUDBGAlpha = 0.4,
    SpeedPrecision = 0,
    ShowPitchText = true,
    ShowHeadingText = true,
    ShowCoordinatesText = true,
    ShowSpeedText = true,
    ShowSpeedPrefix = true,
    LockPosition = true,
    FrameScale = 0.72307443618774,
    PitchGraphicsScale = 17.448093414307,
    FontScale = 20.242206573486,
    FramePoint_point = "TOP",
    FramePoint_relativeTo = false,
    FramePoint_relativePoint = "TOP",
    FramePoint_xOfs = 14.663482434858,
    FramePoint_yOfs = -76.521231176252,
    WorldmapModifier = "S",
    RemoveWaypointDistance = 15,
    QuestHelper = false,
    --[===[@debug@
    EnableDebugWaypoints = false,
    EnableDebugQuestHelper = false,
    --@end-debug@]===]
  }
}
